#ifndef _RDBQUERY_H_
#define _RDBQUERY_H_

// A query is on a join of relations and has a particular adornment.
// A join maintains the set of relations, and any join conditions between
// those columns.  An adornment states which attributes will have selection
// criteria on them when a cursor is opened on the query.

// Idea:  seperate "compilable" to make it only a marker for notification
// of invalidation.  basically it would be a call back stub.  then each
// implementor class could inherit from it if it needs the invalidation.
// ...

#include "RdbCompilableDML.h"
#include "RdbJoin.h"
#include "RdbItem.h"
#include "RdbCursor.h"
#include "String.h"
#include "BitVector.h"
#include "Boolean.h"



class RdbQuery
: public RdbCompilableDML
{
  public:  

    RdbQuery(const RdbJoin& newJoin, const BitVector& newAdorn );

    // should this be implementd?
    // RdbQuery( RdbRelation& newRel, BitVector& newAdorn );
    
    virtual ~RdbQuery();

    int Arity() const;		// number of attributes in query

    Boolean ColumnIsBound(int i) const; // is column i expected to be known?

    const BitVector& Adornment() const; // the adornment of this query

    String Schema() const;	// schema text

    String RelName(int i) const; // i'th relation name

    String ColName(int i) const; // i'th attribute name

    String QualifiedColName(int col) const; // relationName.columnName

    String VarQualifiedColName(int col) const; // tupleVar.columnName

    ErrorCode GetNext(RdbCursor& cursor, RdbItemList& result);
				// get next result from cursor
				// called by cursor.GetNext()

    virtual void const* Implementor() const = 0;
				// access to implementor class (child)
				// returns this.

  protected:

    void CreateStmt(const String varPlaceHolder = questionMark,
		    const int addSequence = 0);
				// creates an ansi standard select statement
				// with a specified variable place holder,
				// adding a sequence number to the variable
				// if requested.

    void CreateStmt(const RdbItemList& bindings); 
				// creates an ansi standard select statement
				// with specified bindings rather than 
				// variables.

    String CreateStmtWithoutConditions();
				// returns the select statement with the
				// join conditions, but no selection 
				// conditions.

    void* GetCursorId(RdbCursor& cursor);
				// friend of RdbCursor, to allow
				// access to the id from all derived
				// queries, but nobody else.

    void SetCursorId(RdbCursor& cursor, void* newId);
				// friend of RdbCursor, to allow
				// updates to the id from all derived
				// queries, but nobody else.

    // these actually perform the operation in the backend
    
    virtual ErrorCode VOpen(RdbCursor& cursor) = 0; 

    virtual ErrorCode VClose(RdbCursor& cursor) = 0;

    virtual ErrorCode VGetNext(RdbCursor& cursor, RdbItemList& result) = 0;

  private:

    RdbJoin join;		// the relation set to query

    BitVector adorn;		// the attrs. that will be known at open

    int arity;			// total number of attributes in result

    int SetFocus();		// set's the focus to this database

    ErrorCode Open(RdbCursor& cursor); // open cursor on this query
				       // called by cursor.Open()

    ErrorCode Close(RdbCursor& cursor); // close this cursor
					// called by cursor.Close()

    RdbQuery(const RdbQuery&);	// na

    RdbQuery& operator =(const RdbQuery&); // na

    friend class RdbCursor;	// for open/close

};


#include "RdbQuery.i"

#endif // _RDBQUERY_H_
