// *************************************************************************
//    CArg.h
//
//    The Coral Arg - All Arguments in Coral can (or will soon) be 
//                    put inside of these.
//
// *************************************************************************

#ifndef _CARG_H
#define _CARG_H

#include <stddef.h>
#include <iostream.h>

#include "HashVal.h"
#include "StrRefTable.h"
#include "CArgUtil.h"
#include "Entity.h"

class  Cor_CArgStack;
class  Cor_TupleStack;

// *************************************************************************

#ifdef   NO_INLINE_CARG
#    define    INLINE        /* Nothing */
#else
#    define    INLINE    inline
#endif

// *************************************************************************

class  Cor_CArg;

extern Cor_CArg   *Cor_TheDontCareArg;

typedef int  Variable;     // This is distinct from VarArg!
typedef int  VarFunc(Cor_CArg *var, void *data);
typedef void ExprFunc(Cor_CArg&, const Cor_CArg&, const Cor_CArg&);

class  Cor_BindEnv;
class  Cor_BitVector;

class  Cor_ArgList;
class  Cor_Relation;
class  Cor_StorageRelation;
class  Cor_Term;
class  Cor_TermLink;
class  Cor_Tuple;

// *************************************************************************

enum CArgKind {

    EMPTY_ARG,      // Right after initialization, no real value

    INT_ARG,        // An integer, stored directly in the union
    DOUBLE_ARG,     // A double, stored directly in the union

    SYMBOL_ARG,     // A string, stored directly in the union
    TINY_SYM_ARG,   // For tiny strings

    CHUNK_ARG,      // A pointer to a chunk of memory (for the C++ Interface)

    FUNCT_ARG,      // A functor term

    VAR_ARG,        // A variable
    REL_ARG,        // A relation
    DONT_CARE_ARG,  // The don't care Arg
    PARAMETER_ARG,  // The parameter Arg (for the C++ Interface)

      // Lists
    NIL_ARG,        // nil
    LIST_ARG,       // A (non-nil) list

      // ------------------------------
      // Arithmetic Expressions
    ADD_OP_ARG,
    SUB_OP_ARG,
    MULT_OP_ARG,
    DIV_OP_ARG,
    IDIV_OP_ARG,
    MOD_OP_ARG,

    LSHFT_OP_ARG,
    RSHFT_OP_ARG,

    BT_AND_OP_ARG,
    BT_OR_OP_ARG,
    BT_NOT_OP_ARG,
    BT_XOR_OP_ARG,

      // ------------------------------
      // The various types of Grouping Args
      // If you change this ORDERING, or add one, make sure
      // to update the builtin_ops[] array in Gram.y as well.

    SUM_GRPING_ARG,       // sum(<X>)
    PROD_GRPING_ARG,      // prod(<X>)
    CNT_GRPING_ARG,       // count(<X>)
    AVG_GRPING_ARG,       // avg(<X>)
    STDDEV_GRPING_ARG,    // stddev(<X>)

    SUM_DGRPING_ARG,      // sum_distinct(<X>)
    PROD_DGRPING_ARG,     // prod_distinct(<X>)
    CNT_DGRPING_ARG,      // count_distinct(<X>)
    AVG_DGRPING_ARG,      // avg_distinct(<X>)

    MIN_GRPING_ARG,       // min(<X>)
    MAX_GRPING_ARG,       // max(<X>)
    SET_GRPING_ARG,       // distinct(<X>)
    ANY_GRPING_ARG,       // any(<X>)
    NO_AGG_GRPING_ARG,    // <X>

    USR_DEFD_GRPING_ARG,  // User-defined grouping

      // ------------------------------
      // Aggregate Result (The result from some types of groupings)
    AGG_RESULT_ARG,
      // For more complex Aggregates (not in use yet)
    CMPLX_AGG_RESULT_ARG

      // ------------------------------
};

  // boolean is going to be in the C++ standard eventually. Til then,
enum truth {
    NO = 0,
    YES
};

// *************************************************************************

class Cor_CArg {

  private:

    CArgKind     my_arg_kind;

    union /* Auxilary Data */ {
        HashVal      _hash;        // Hash for items that are hashed
        int          _chk_subsum;  // For Grouping
        Cor_CArg    *_agg_num;     // A number for aggregation (??)
    };

    union /* The Main Data */ {

      double           _doub;
      int              _int;
      int              _var_num;

      struct sym_st {         // Symbols
          char        *_symbol;
          int         *_str_ref_cnt;
      } sym;

                  // Tiny Symbols
      char         _symbuf[sizeof(double)];

      struct func_st {        // Functor Terms
          Cor_CArg     *_functor;
          Cor_ArgList  *_args;
      } func;

      struct list_st {        // Lists
          unsigned short   _num_items;    // Number of top-level items
          Cor_CArgStack   *_items;        // List items are stored reversed
      } list;

      struct expr_st {        // Expressions
          Cor_CArg    *_larg;
          Cor_CArg    *_rarg;
      } expr;

      struct grp_st  {        // For Grouping
          Cor_CArg    *_nested_args;
          int          _seq_num;    // sequence number within current clause
      } grp;

      struct agg_st  {        // Aggregates
          int                    _agg_count;
          Cor_StorageRelation   *_dist_rel;
      } agg;

      struct cmplx_agg_st  { // More complex Aggregates (std deviation)
          int                    _agg_count;
          Cor_CArgStack         *_agg_stk;   // A stack of values
      } cmplx_agg;

      void            *_chunk;
      Cor_Relation    *_rel;

      Cor_CArg              *nxt_free;  // For free-list of CArgs
    };

    void print_binary_expr(Cor_BindEnv *context, ostream *file, char *op) const;
    void print_expr_qv(Cor_BindEnv *context, ostream *file) const;

    void sprint_grping(char *str, int *pos, Cor_BindEnv *env,
                char *start, char *end) const; 
    void print_grping(Cor_BindEnv *context, ostream *file,
                char *start, char *end) const; 
    void print_grping_qv(Cor_BindEnv *context, ostream *file,
                char *start, char *end) const; 

    void simplify_expr(Cor_CArg& carg, // result = the simplified carg!
          Cor_BindEnv *context, 
          Cor_TermLink *&renamed_vars,
          Cor_BindEnv *const_env, 
          Cor_BindEnv *rename_only_env,
          ExprFunc    *efunc);


    HashVal functor_hash(Cor_BindEnv *context);
    HashVal list_hash(Cor_BindEnv *context);

    int     compute_isConstant() const;

    int functor_equals(Cor_CArg *carg);
    int list_equals(Cor_CArg *carg);

  public:
#ifdef    REUSE_MEMORY
    void* operator new(size_t size) {
        return(Coral_Free->acquire(size));
    }

    void  operator delete(void *trash, size_t size) {
        Coral_Free->dispose((Cor_Ent_Element*) trash, size);
    }
#endif

    Cor_CArg(void);                    // Empty
    Cor_CArg(const Cor_CArg& a);
    Cor_CArg(int n);                   // Numbers
    Cor_CArg(double d);                // Numbers
    Cor_CArg(const char *name);        // Symbols
    Cor_CArg(Cor_Relation *relation);  // Relations

    Cor_CArg(CArgKind kind);           // Whatever type is given in kind

    Cor_CArg(Cor_CArgStack *stk);     // Lists
    Cor_CArg(Cor_CArg *car, Cor_CArg *cdr);

    Cor_CArg(CArgKind kind, Cor_CArg *larg, Cor_CArg *rarg);  // Expressions

    Cor_CArg(Cor_CArg& functor, int arity);  // Functors

          // Should only be used internally (because of ref cnting)
    INLINE ~Cor_CArg();

    INLINE void init();      // Reinitializes, does not free up any space
    void free_mem();  // Frees up any space this CArg has, then does an init()

    // The reference counting stuff
    // ----------------------------

      // Duplicates 'this', using reference counting if possible
    Cor_CArg *dup();

      // Duplicate a list, don't use ref counting
    Cor_CArg *list_dup();

      // Same as "delete"
    void      destroy() { delete this; }

    // ----------------------------

    Cor_CArg& operator= (const Cor_CArg& a);
    Cor_CArg& operator= (int i);
    Cor_CArg& operator= (double d);

    int operator ==(const Cor_CArg& c) const;

    void set_kind(CArgKind kind);

    void set_chunk(void *chnk);
    void set_num(int ival);
    void set_num(double dval);
    void set_var(int x);

    void set_grping(int arg_cnt);
    void set_grping(Cor_CArg *nested_args, int seq_num=0,
                    CArgKind kind=NO_AGG_GRPING_ARG);

    void set_grp_subsum(int chk_subsum);
    void set_grp_seqnum(int seqnum);
    void set_grp_kind(CArgKind kind);

      // test if two CArgs are equal,
    int equals(const Cor_CArg& carg) const;
    int equals(Cor_CArg *carg);
    int equals(char *str) const;

      // Test if the symbols stored in these items are the same
      // (the two items need not be the same type)
    int nm_equals(const Cor_CArg& arg) const;

    CArgKind Kind() const;

    double NumDval()   const;
    int NumIval()      const;
    int isConstant()   const;

    INLINE int is_empty()     const;   // empty arg or Other_arg (0)
    INLINE int is_num()       const;
    INLINE int is_int()       const;
    INLINE int is_short()     const;
    INLINE int is_long()      const;
    INLINE int is_float()     const;
    INLINE int is_double()    const;

    INLINE int is_string()    const;
    INLINE int is_symbol()    const;

    INLINE int is_chunk()     const;
    INLINE int is_param()     const;

    INLINE int is_functor()   const;
    INLINE int is_var()       const;
    INLINE int is_rel()       const;
    INLINE int is_dont_care() const;

    INLINE int is_list()      const;
    INLINE int is_cons()      const;
    INLINE int is_nil()       const;

    int is_expr()      const;
    int is_grouping()  const;

    int same_kind(Cor_CArg *carg) const;

      // compares two numbers. assume they both are.
    int compare_num(const Cor_CArg& arg2) const;

      // compares _any_ two args, return <0, =0, or >0
      // depending on relation being object and "that".
      // The result is arbitrary (but consistent),
      // when the types are incompatible.
    int compare(Cor_CArg *that) const;
    int compare_decr(Cor_CArg *that) const {
        int cmp = compare(that);
        if (cmp) {
            if (cmp < 0)
                cmp =  1;
            else
                cmp = -1;
        }
        return(cmp);
    }

      // Returns the hash value for a CArg of any type
    HashVal hash(Cor_BindEnv *context = 0) const;

    void simplify(Cor_CArg& carg, // result = the simplified carg!
          Cor_BindEnv *context, 
          Cor_TermLink *&renamed_vars,
          Cor_BindEnv *const_env, 
          Cor_BindEnv *rename_only_env);

      // Special case for simplifying yourself, return value is 0
      // if variables still remain, 1 if new value is a constant.
    int simplify_me(Cor_BindEnv *context);

    int simplify_me_expr(Cor_BindEnv *context, ExprFunc *efunc);

      // returns the length of the string, if symbol arg. no err check
    INLINE int   symlen() const; 
    INLINE char *symbol() const;
    INLINE HashVal symhash() const;

    INLINE char *functor() const;
    INLINE Cor_CArg *cfunctor() const;

    Variable max_vars();
    void var_list(Cor_IntStack *var_stk);
    int  scan_vars(void *data, VarFunc *func);

      // Stream printing
    friend ostream& operator<< (ostream& out, const Cor_CArg& carg);

      // Would the given string require quotes
    int  need_quotes(char *str) const;

    void print(Cor_BindEnv *context, ostream *file) const;
    void print_qv(Cor_BindEnv *context, ostream *file) const;

    void print_dump(Cor_BindEnv *context, ostream *file) const;
    void sprint(char *str, int *pos, Cor_BindEnv *context = NULL) const;
    void printon(ostream *file) const;

    void output() const;

    void *chunk();

      // Functors
    Cor_ArgList& args();
    int arity();

      // Variables
    int  var_num() const;

    Cor_Relation *rel() const;

      // Lists
    Cor_CArg *Cor_CArg::car() const;
    void Cor_CArg::get_cdr(Cor_CArg& cdr) const;
    Cor_CArg *Cor_CArg::cdr() const;    // implies duplication

      // Remove the recursive list structure resulting from evaluation
    Cor_CArg *Cor_CArg::normalize_list();

    int  num_items();
    int  incr_num_items(int chg=1);
    Cor_CArgStack *items();

        // UnMake a list of items into a tuple.
        // ('make_list()' is in the ArgList class.)
    Cor_Tuple *unmake_list();
        // UnMake a list of lists
    int unmake_lists(Cor_TupleStack *stk);

      // Substitute the parameters in.
    int substitute_params( Cor_CArgStack *params );

      // Expressions
    CArgKind  Cor_CArg::op() const;
    Cor_CArg *Cor_CArg::larg() const;
    Cor_CArg *Cor_CArg::rarg() const;

      // Grouping
    Cor_CArg *nested_args();
    int grp_arity() const;
    int update_tuple(Cor_Tuple *old_tup, Cor_CArg val,
                    int argnum, Cor_BindEnv *context);

      // Aggregates
    Cor_StorageRelation *dist_rel();

    // For debugging use
    void show_hash() { cout << _hash << " "; }
    size_t show_internals() const;
    char  *kind_str() const;
};


// *************************************************************************

const  Cor_CArg Cor_NullCArg;

const  Cor_CArg Cor_ZeroCArg(0);
const  Cor_CArg Cor_MinusOneCArg(-1);
const  Cor_CArg Cor_OneCArg(1);

extern Cor_CArg *Cor_DontCareSymbol;

// *************************************************************************

  // Include the inlined functions
#include "CArg.i"

#endif _CARG_H
