// *************************************************************************
//  Collection.C
// *************************************************************************

#include "Collection.h"

#include "Rule.h"
#include "Literal.h"
#include "Stack.h"
#include "SymTable.h"
#include "Globals.h"
#include "Debug.h"


// *************************************************************************

    // DEFAULT size of an array of Collections
#define    DEFAULT_SIZE    64

// *************************************************************************
Cor_Collection::Cor_Collection(int init_size, int init_hash_on)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    length  = 0; 
    hash_on = init_hash_on;

    if (init_size)
        size = init_size;
    else
        size = COR_DEFLT_COLLCTN_SIZE;

    rules = new Cor_Clause* [size];

    for (int j=0; j < size; j++)
        rules[j] = NULL;

    if (hash_on)
        table = new Cor_SymTable();
    else
        table = NULL;

    param_slots = NULL;
}

// *************************************************************************
Cor_Collection::~Cor_Collection()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
/*
    for (int j=0; j < length; j++) {
        delete rules[j];
    }
*/

    delete [] rules;
    delete table;

    delete param_slots;
}

// *************************************************************************
// If hashing is on, return the first Clause that has this name,
// the others are obtained by following the Clause's link field.
// If hashing is off, this always returns NULL (not found).
Cor_Clause *Cor_Collection::get_match(Cor_CArg& name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (! hash_on)
        return(NULL);

    return((Cor_Clause *) table->lookup(name));
}

// *************************************************************************
// To turn hashing on and off (this is idempotent)
int Cor_Collection::use_hash_table(int on)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j;

    if ((on && hash_on) || (!on && !hash_on))
        return(hash_on);

    hash_on = on;

    if (hash_on) {

        // Turn on hash table
        table = new Cor_SymTable();

        for (j=0; j < length; j++) {

            Cor_Clause *rule = rules[j];

            if (! rule)
                continue;
            rule->link = NULL;

            Cor_Clause *prev = (Cor_Clause*) table->lookup(rule->head->pred);

            if (prev) {

                    // Add it on the end of the list
                for ( ; prev->link != NULL; prev = prev->link)
                    ;    // Nothing

                prev->link = rule; 

            } else {
                table->insert(rule->head->pred, (Pointer) rule);
            }
        }

    } else {

        // Turn off hash table
        for (j=0; j < length; j++)
            (rules[j])->link = NULL;

        delete table;
        table = NULL;
    }

    return(hash_on);
}

// *************************************************************************
// push() returns the position of the inserted rule
int Cor_Collection::push(Cor_Clause *rule)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j;

    if (Cor_exEnv->C_exec_mode == COR_DOING_IMPERATIVE) {
        for (j=0; j < size; j++)
            if (rules[j] == rule)
                return(j);
    }

      // Check if the array needs to be larger
    if (length >= size) {
        Cor_Clause  **old_rules = rules;
        size = size + size;

        rules = new Cor_Clause* [size];

        for (j=0; j < size; j++)
            rules[j] = NULL;

        for (j=0; j < length; j++)
            rules[j] = old_rules[j];

        delete [] old_rules;
    }

    rules[length++] = rule;
    rule->link = NULL;

      // Check if we need to add it to a list
    if (hash_on) {

        Cor_Clause *prev = (Cor_Clause*) table->lookup(rule->head->pred);

        if (prev) {

                // Add it on the end of the list
            for ( ; prev->link != NULL; prev = prev->link)
                ;    // Nothing

            prev->link = rule; 

        } else {
            table->insert(rule->head->pred, (Pointer) rule);
        }
    }

    return(length-1);
}

// *************************************************************************
// deletes all stored rules and re-initializes the structure
void Cor_Collection::empty()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    for (int j=0; j < length; j++) {
        delete rules[j];
        rules[j] = NULL;
    }

    length = 0;

      // Clear out the hash table
    if (hash_on)
        table->clear();
}

// *************************************************************************
// *************************************************************************
// push() returns the position of the inserted collection 
int  Cor_CollectStack::push(Cor_Collection *x)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (length >= data_size) {
        if (data_size == 0) {
            data_size = DEFAULT_SIZE;
            data = new Cor_Collection*[data_size];
        } else {
            int len = data_size;
            data_size += data_size;
            Cor_Collection **tmp_data = data;
            data = new Cor_Collection*[data_size];
            for (int i=0; i<len; i++)
                data[i] = tmp_data[i];
            delete [] tmp_data;
        }
    }
 
    data[length++] = x;

    return(length-1);
}

// *************************************************************************
//  Empty out the array - delete each element
void Cor_CollectStack::empty()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    for (int j=0; j < length; j++) {
        delete data[j];
        data[j] = NULL;
    }

    length = 0;
}

#undef    DEFAULT_SIZE
// *************************************************************************
// After a few del_ith's, your array will have holes, this is
// bad, so this function compacts it.
// It returns the number of holes removed
int Cor_Collection::squash()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j, hole;
    int nholes = 1;
       
      // find the first hole 
    for (hole=0; hole < length; hole++)
        if (! rules[hole])
            break;

      // No squashing needed
    if (hole == length)
        return(0);

      // move all the holes down to the right end of the array
    for (j=hole+1; j < length; j++) {

        if (rules[j]) {

            rules[hole] = rules[j];
            rules[j] = NULL;

              // Move to the next hole (might be the one just created)
            hole++;

        } else {
            nholes++;
        }
    }

    length -= nholes;
    return(nholes);
}

// *************************************************************************
