// *************************************************************************
// AsynSocketRelSolver.C
// solver function for builtin: asyn_socket_rel
// *************************************************************************

// user_command: asyn_socket_rel
// syntax: asyn_socket_rel(local_name, arity (or func-arg), host, port,
//                         work_space, relation name)
//     or asyn_socket_rel(local_name(_,_,....), host, port,
//                        work_space, relation name)
// description: asyn_socket_rel defines a local relation which is stored in
//     a remote host, accessed via a socket port, in a work_space, with a
//     remote name.
// error_handling: the following situation will cause an error msg echoed 
//     to the user, not when 
//      1) unknown host, unknown port, connection lost
//      2) wrong workspace, relation name, argument number 


#include "Globals.h"
#include "CoralError.h"
#include "BuiltinRelation.h"
#include "TupleIterator.h"
#include "AsynSockRelation.h"

#define FAIL_RETURN  do {iterator.set_no_match(); return(NULL);} while(0)

// *************************************************************************
Cor_BindEnv *AsynSocketRelSolver(Cor_BuiltinRelation& rel,
        Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef ASYN_SOCK_REL

    char *local_name, *host, *relation_name, *work_space;
    int arity, port;
   
    if ((iterator.arg_list.count() != 6) && (iterator.arg_list.count() !=5))  {
        CORAL_error(COR_BAD_ARITY,"5 or 6 arguments to asyn_socket_rel needed",
                "AsynSocketRelSolver");
        FAIL_RETURN;
    }
    
    if (iterator.arg_list.count() == 6) {
        Cor_Term term1 (iterator.arg_list[0], iterator.bindenv);
        Cor_deep_dereference_term (term1);
        Cor_Term term2 (iterator.arg_list[1], iterator.bindenv);
        Cor_deep_dereference_term (term2);
        Cor_Term term3 (iterator.arg_list[2], iterator.bindenv);
        Cor_deep_dereference_term (term3);
        Cor_Term term4 (iterator.arg_list[3], iterator.bindenv);
        Cor_deep_dereference_term (term4);
        Cor_Term term5 (iterator.arg_list[4], iterator.bindenv);
        Cor_deep_dereference_term (term5);
        Cor_Term term6 (iterator.arg_list[5], iterator.bindenv);
        Cor_deep_dereference_term (term6);

        if (!term1.expr.is_symbol()) {
            CORAL_error(COR_BAD_ARGLIST, "Local name must be a symbol",
                    "AsynSocketRelSolver");
            *(Cor_exEnv->err_strm) << term1 << endl;
            FAIL_RETURN;
        }

        if (term2.expr.is_functor()) 
            arity = term2.expr.arity();
        else if (term2.expr.is_num())
            arity = term2.expr.NumIval();
        else {
            CORAL_error(COR_BAD_ARGLIST, "Arity must be a number or func-arg",
                    "AsynSocketRelSolver");
            *(Cor_exEnv->err_strm) << term2 << endl;
            FAIL_RETURN;
        }
        
        if (!term3.expr.is_symbol()) {
            CORAL_error(COR_BAD_ARGLIST, "Host must be a symbol",
                    "AsynSocketRelSolver");
            *(Cor_exEnv->err_strm) << term3 << endl;
            FAIL_RETURN;
        }

        if (!term4.expr.is_num()) {
            CORAL_error(COR_BAD_ARGLIST, "Port must be a number",
                    "AsynSocketRelSolver");
            *(Cor_exEnv->err_strm) << term4 << endl;
            FAIL_RETURN;
        }

        if (!term5.expr.is_symbol()) {
            CORAL_error(COR_BAD_ARGLIST, "Workspace must be a symbol",
                    "AsynSocketRelSolver");
            *(Cor_exEnv->err_strm) << term5 << endl;
            FAIL_RETURN;
        }

        if (!term6.expr.is_symbol()) {
            CORAL_error(COR_BAD_ARGLIST, "Relation name must be a symbol",
                    "AsynSocketRelSolver");
            *(Cor_exEnv->err_strm) << term6 << endl;
            FAIL_RETURN;
        }
       
        local_name    = term1.expr.symbol();
        host          = term3.expr.symbol();
        port          = term4.expr.NumIval();
        work_space    = term5.expr.symbol();
        relation_name = term6.expr.symbol();

    } else {
        Cor_Term term1 (iterator.arg_list[0], iterator.bindenv);
        Cor_deep_dereference_term (term1);
        Cor_Term term2 (iterator.arg_list[1], iterator.bindenv);
        Cor_deep_dereference_term (term2);
        Cor_Term term3 (iterator.arg_list[2], iterator.bindenv);
        Cor_deep_dereference_term (term3);
        Cor_Term term4 (iterator.arg_list[3], iterator.bindenv);
        Cor_deep_dereference_term (term4);
        Cor_Term term5 (iterator.arg_list[4], iterator.bindenv);
        Cor_deep_dereference_term (term5);
        
        if (!term1.expr.is_functor()) {
            CORAL_error(COR_BAD_ARGLIST, "Local name must be a functor",
                    "AsynSocketRelSolver");
            *(Cor_exEnv->err_strm) << term1 << endl;
            FAIL_RETURN;
        }

        if (!term2.expr.is_symbol()) {
            CORAL_error(COR_BAD_ARGLIST, "Host must be a symbol",
                    "AsynSocketRelSolver");
            *(Cor_exEnv->err_strm) << term2 << endl;
            FAIL_RETURN;
        }

        if (!term3.expr.is_num()) {
            CORAL_error(COR_BAD_ARGLIST, "Port must be a number",
                    "AsynSocketRelSolver");
            *(Cor_exEnv->err_strm) << term3 << endl;
            FAIL_RETURN;
        }

        if (!term4.expr.is_symbol()) {
            CORAL_error(COR_BAD_ARGLIST, "Workspace must be a symbol",
                    "AsynSocketRelSolver");
            *(Cor_exEnv->err_strm) << term4 << endl;
            FAIL_RETURN;
        }

        if (!term5.expr.is_symbol()) {
            CORAL_error(COR_BAD_ARGLIST, "Relation name must be a symbol",
                    "AsynSocketRelSolver");
            *(Cor_exEnv->err_strm) << term5 << endl;
            FAIL_RETURN;
        }
       
        local_name    = term1.expr.functor();
        arity         = term1.expr.arity();
        host          = term2.expr.symbol();
        port          = term3.expr.NumIval();
        work_space    = term4.expr.symbol();
        relation_name = term5.expr.symbol();

    }
 
    ASSERT( local_name && *local_name );
          
#ifdef DEBUG_ASYN_SOCK_REL
    *(Cor_exEnv->out_strm) << "local_name = " << local_name << endl;
    *(Cor_exEnv->out_strm) << "arity = " << arity << endl;
    *(Cor_exEnv->out_strm) << "host = " << host << endl;
    *(Cor_exEnv->out_strm) << "port  = " << port << endl;
    *(Cor_exEnv->out_strm) << "work_space = " << work_space << endl;
    *(Cor_exEnv->out_strm) << "relation_name = " << relation_name << endl;
#endif
        
    Cor_Relation *new_rel;
    Cor_CArg loc_name(local_name);

    new_rel = Cor_find_external_relation(loc_name,arity);

    if (new_rel) 
        *(Cor_exEnv->err_strm)
             << "Error: there is already a local relation named "
             << local_name << endl;
    else {
        Cor_AsynSockRelation *sock_rel
            = new Cor_AsynSockRelation(local_name, arity, host, port,
                work_space, relation_name);
        Cor_CurDB->RelationTable->insert(loc_name, sock_rel);
        *(Cor_exEnv->err_strm) << "asyn_socket relation " << local_name
                               << " created.\n";
    }
#else
    *(Cor_exEnv->err_strm)
        << "The command '" << *Cor_AsynSocketRelSymbol << "' is unavailable.\n"
        << "Sockets are not compiled into this version of Coral.\n";

    iterator.set_no_match();
#endif // ASYN_SOCK_REL

    return(iterator.bindenv);
}

// *************************************************************************
